/*
 * Mess with the sys_call_table in kernel memory.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/unistd.h>
#include <linux/kallsyms.h>
#include <linux/sched.h>
/* #include <asm/cacheflush.h> */

MODULE_AUTHOR("Juergen Schoenwaelder");
MODULE_LICENSE("Dual BSD/GPL");
MODULE_DESCRIPTION("Mess with the sys_call_table in kernel memory.");

/*
 * For XEN, we disable / enable page protection. This is an ugly hack,
 * don't do this for production. You have been warned.
 */

#define CR0_BIT_WRITE_PROTECT 0x10000

static void gpf_disable(void)
{
    unsigned long cr0;

    cr0 = read_cr0();
    cr0 &= ~CR0_BIT_WRITE_PROTECT;
    write_cr0(cr0);
}

static void gpf_enable(void)
{
    unsigned long cr0;

    cr0 = read_cr0();
    cr0 |= CR0_BIT_WRITE_PROTECT;
    write_cr0(cr0);
}

/*
 * The page holding the sys_call_table is write protected. So we have
 * to change that before we can mess with it by looking up the page
 * table entry and changing the page table bits.
 */

static void set_addr_rw(unsigned long addr)
{
    unsigned int level;
    pte_t *pte = lookup_address(addr, &level);

    if (pte->pte &~ _PAGE_RW) pte->pte |= _PAGE_RW;
}

static void set_addr_ro(unsigned long addr)
{
    unsigned int level;
    pte_t *pte = lookup_address(addr, &level);

    pte->pte = pte->pte &~_PAGE_RW;
}

/*
 * Try to locate the sys_call_table. On some kernels, we can simply
 * lookup the symbol. On other kernels, we may have to serach for
 * it using some heuristics (which is dangerous of course).
 */

static unsigned long *p_sys_call_table = NULL;

static int lookup_syscalltable(void)
{
    unsigned long *sct;

    sct = (unsigned long *) kallsyms_lookup_name("sys_call_table");
    if (! sct) {
        return -ENOENT;
    }

    p_sys_call_table = sct;
    return 0;
}

asmlinkage int (*original_call) (const char*, int, int);

asmlinkage int our_sys_open(const char* file, int flags, int mode)
{
    printk(KERN_ALERT "[%s] open('%s') by task %i (%s)\n",
	   __this_module.name, file, current->pid, current->comm);
    return original_call(file, flags, mode);
}

static int __init syscall_init(void)
{
    int rc;

    rc = lookup_syscalltable();
    if (rc) {
        return rc;
    }
    printk(KERN_DEBUG "[%s] sys_call_table found at %lx\n", 
	   __this_module.name, (unsigned long) p_sys_call_table);
    printk(KERN_ALERT "[%s] replacing sys_open with our_sys_open\n",
	   __this_module.name);

    original_call = (void *) p_sys_call_table[__NR_open];
    gpf_disable();
    set_addr_rw((unsigned long) p_sys_call_table);
    p_sys_call_table[__NR_open] = (unsigned long) our_sys_open;
    gpf_enable();
    return rc;
}

static void __exit syscall_exit(void)
{
    printk(KERN_ALERT "[%s] replacing our_sys_open with sys_open\n",
	   __this_module.name);

    gpf_disable();
    p_sys_call_table[__NR_open] = (unsigned long) original_call;
    set_addr_ro((unsigned long) p_sys_call_table);
    gpf_enable();
}

module_init(syscall_init);
module_exit(syscall_exit);
